#ifdef ESP32
#include "WiFi.h"
#include <SPIFFS.h>
#include <ESPmDNS.h>
#else
#include "ESP8266WiFi.h"
#include <FS.h>
#include <ESP8266mDNS.h>
#endif
#include "ESPAsyncWebServer.h"
#include <movingAvg.h>

movingAvg distanceAvg(10);  // Take average of 10 readings for ultrasonic sensor
movingAvg frAvg(10);        // Take average of 10 readings for flow sensor

// Pin assignments
const int trigPin = D5;  // Ultrasonic Trigger pin
const int echoPin = D6;  // Ultrasonic Echo pin
const int flowSensorPin = D1; // Flow sensor signal pin

// Sensor Variables:
float CalibrationFactor = 7.5; //This is the calibration factor for my flow rate sensor you can adjust it for better results

// Tank Variables for Ultrasonic sensor
int minHeight = 25;   // It is the distance from ultrasonic sensor to the water when tank is full
int maxHeight = 115;  // It is the distance from ultrasonic sensor to the bottom of the tank or min water level

// WiFi network credentials
const char* ssid = "YourWifiSSID";         // Add your Wi-Fi SSID here
const char* password = "YourWifiPassword"; // Add your Wi-Fi password here

// Create AsyncWeb object on port 80
AsyncWebServer server(80);

// Variables for flow rate sensor
volatile int pulseCount = 0;
float flowRate = 0.0;
unsigned long oldTime = 0, readingInterval = millis();
int percentage;
float dAvg;

// If we use any placeholder in html code then this function is useful
String processor(const String& var) {
  return String();
}

// Checks if motion was detected, sets LED HIGH and starts a timer
ICACHE_RAM_ATTR void pulseCounter() {
  pulseCount++; // Increment pulse count whenever a pulse is detected
}

void setup() {
  // Serial port for debugging purposes
  Serial.begin(115200);
  if (!SPIFFS.begin()) {
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }
  pinMode(D4, OUTPUT);
  digitalWrite(D4, LOW);

  distanceAvg.begin();
  frAvg.begin();

  WiFi.begin(ssid, password);
  Serial.print("Connecting..");
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print(".");
    delay(1000);
  }
  Serial.println();
  Serial.println("connected");
  digitalWrite(D4, HIGH);

  // Initialize mDNS
  if (MDNS.begin("watertank")) {  // Replace with your preferred hostname
    Serial.println("mDNS responder started");
    Serial.println("You can access the server at http://watertank.local");
  } else {
    Serial.println("Error setting up mDNS responder!");
  }
  // Set up ultrasonic sensor pins
  pinMode(trigPin, OUTPUT);
  pinMode(echoPin, INPUT);

  // Set up flow sensor pin
  pinMode(flowSensorPin, INPUT);
  attachInterrupt(digitalPinToInterrupt(flowSensorPin), pulseCounter, FALLING); // Interrupt for flow sensor

  // Initialize flow rate calculation
  oldTime = millis();

  //Calling the files from SPIFFS memory of esp32
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/index.html", String(), false, processor);
  });

  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/style.css", "text/css");
  });
  // Serve the JavaScript file
  server.on("/script.js", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(SPIFFS, "/script.js", "application/javascript");
  });
  // sending the required data on paths /distance and /frate
  server.on("/distance", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(200, "text/plain", String(percentage));
  });

  server.on("/frate", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send(200, "text/plain", String( frAvg.reading(flowRate)));
  });
  // Start server
  server.begin();
}

void loop() {
#ifdef ESP8266
  MDNS.update();  // Add this to keep mDNS running smoothly
#endif

  if (WiFi.status() != WL_CONNECTED)  //If WiFi network stopped working
  {
    WiFi.reconnect();       //trying reconnecting the same previous network
    while (WiFi.status() != WL_CONNECTED) {
      Serial.print(".");
      delay(1000);
    }
    Serial.println();
    Serial.println("connected");
  }

  if ( millis() - readingInterval > 1000) //reading the sensor values with 1 second interval
  {
    readingInterval = millis();
    distance();
    flowrate();
  }

}

// Function to measure distance using ultrasonic sensor
String distance() {
  long duration;
  float distanceCm;

  // Send a 10us pulse to trigger the sensor
  digitalWrite(trigPin, LOW);
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH);
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);

  // Measure the time for the echo to return
  duration = pulseIn(echoPin, HIGH);

  // Convert time to distance in cm
  distanceCm = duration * 0.034 / 2;

  if (distanceCm >= minHeight && distanceCm <= maxHeight)
  {
    percentage = map(distanceAvg.reading(distanceCm), minHeight, maxHeight, 0, 100);

    if (percentage < 0 ) percentage = 0;
    else if (percentage > 100 ) percentage = 100;
    percentage = 100 - percentage;
  }
  Serial.println(percentage);
  return String(percentage);
}

// Function to calculate flow rate using flow rate sensor
String flowrate() {
  unsigned long currentTime = millis();
  unsigned long timeInterval = currentTime - oldTime;

  // Calculate flow rate based on pulse count
  if (timeInterval > 1000) { // Calculate every 1 second
    flowRate = (pulseCount / CalibrationFactor); // Formula for YF-S201 flow sensor (adjust based on your sensor)
    oldTime = currentTime;         // Reset time interval
    pulseCount = 0;                // Reset pulse count for next interval
  }
  Serial.println(flowRate);
  return String(frAvg.reading(flowRate));
}
